% [X,E,A] = ee(Wp,Wn,l,X0[,method,tol,maxit]) Elastic Embedding algorithm
%
% In:
%   Wp, Wn: NxN positive and negative weight matrices.
%   l: Hx1 list of homotopy parameters in increasing order.
%   X0: NxL matrix, initial low-dim coordinates.
%   method: search direction method. Default: 'diagpdH'.
%   tol: minimum distance between consecutive X to keep iterating for each
%      value of l. Default: 1e-3.
%   maxit: maximum number of iterations for each value of l. Default: 100.
% Out:
%   X, E, A: (H+1)x1 cell array where element i+1 is the ith-l-value low-dim
%      coordinates, error curve, and step sizes, respectively.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2010 by Miguel A. Carreira-Perpinan

function [X,E,A] = ee(Wp,Wn,l,X0,method,tol,maxit)

N = size(Wp,1);

% ---------- Argument defaults ----------
if ~exist('method','var') | isempty(method) method = 'diagpdH'; end;
if ~exist('tol','var') | isempty(tol) tol = 1e-3; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 100; end;
% ---------- End of "argument defaults" ----------

X = {X0}; Xold = X0; A = {NaN};
% Ensure Wp and Wn have zero diagonal and are symmetric
Wp = (Wp + Wp')/2; Wn = (Wn + Wn')/2; Wp(1:N+1:N^2) = 0; Wn(1:N+1:N^2) = 0;
% Degree matrix of Wp
Dp = sum(Wp,2); invDp = diag(sparse(Dp.^(-1))); Dp = diag(sparse(Dp));
if issparse(Wp) & issparse(Wn)
  [I,J] = find(triu(Wp,1)|triu(Wn,1));	% Pattern of nonzeros in Wp or Wn
else
  I = []; J = [];
end
[e,sqd,esqd] = ee_error(I,J,Wp,Wn,l(1),X0); E = {e};

for i = 1:length(l)
  
  convcrit = maxit>=1; e = []; a = []; j = 0; alpha = 1;
  while convcrit
  
    WWn = l(i)*Wn.*esqd; DDn = diag(sparse(sum(WWn,2)));
    G = 4*(Dp-Wp-DDn+WWn)*Xold;		% Gradient of E
    switch method			% Search direction for each method
     case 'fp1'
      P = invDp*((Wp+DDn-WWn)*Xold) - Xold;
     case 'fp2'
      P = (Dp+WWn)\((DDn+Wp)*Xold) - Xold;
     case 'fp2-GS'
      P = (Dp+tril(WWn,-1))\((DDn+Wp-triu(WWn,1))*Xold) - Xold;
     case 'fp3'
      P = (DDn+Wp)\((Dp+WWn)*Xold) - Xold;
     case 'Jacobi'
      W = Wp - WWn; D = sum(W,2); P = diag(sparse(D.^(-1)))*(W*Xold) - Xold;
     case 'GS'
      W = Wp - WWn; D = sum(W,2); P = (D-tril(W,-1))\(triu(W,1)*Xold) - Xold;
     case 'graddesc'
      P = -G;
     case 'diagpdH'
      dH = bsxfun(@plus,8*((DDn+WWn)*(Xold.^2)-2*Xold.*(WWn*Xold)),4*diag(Dp));
      P = -G./dH;
    end
    if j==0	% Error
      err = Wp(:)'*sqd(:) + l(i)*Wn(:)'*esqd(:);
      % Although the following is simpler and theoretically identical:
      %   err = Wp(:)'*sqd(:) + sum(diag(DDn));
      % it can cause small numerical differences that cause an infinite loop
      % in the line search (eels.m), so I must compute the error function in
      % exactly the same way each time.
    else
      err = e(j);
    end
    % Line search
    [alpha,Xl,sqd,esqd,el] = eels(I,J,Wp,Wn,l(i),Xold,P,err,G,alpha);
    j = j + 1; e(j) = el; a(j) = alpha;
    convcrit = (norm(Xl-Xold,'fro') > tol*norm(Xold,'fro')) & (j < maxit);
    Xold = Xl;
    
  end
  
  X{i+1} = Xl; E{i+1} = e; A{i+1} = a;
end


% [alpha,X,sqd,esqd,e] = eels(I,J,Wp,Wn,l,X,P,ff,G[,alpha0,rho,c])
% Backtracking line search for EE
%
% Reference: procedure 3.1, p. 41ff in:
%   Nocedal and Wright: "Numerical Optimization", Springer, 1999.
%
% In:
%   I,J: indices of nonzeros in the upper triangular part of Wp+Wn, or empty
%      if they are not sparse.
%   Wp,Wn: as in ee.m.
%   l: current homotopy parameter value.
%   X: NxL matrix, the current iterate.
%   P: NxL matrix, the search direction.
%   ff: value at X of the error function.
%   G: NxL matrix, gradient at X of the error function.
%   alpha0: initial step size. Default: 1.
%   rho: rate of decrease of the step size. Default: 0.8.
%   c: Wolfe condition. Default: 1e-4.
% Out:
%   alpha: step size.
%   X,sqd,esqd,e: new iterate, sqdist(X), exp(-sqd) and error function value.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2009 by Miguel A. Carreira-Perpinan

function [alpha,X,sqd,esqd,e] = eels(I,J,Wp,Wn,l,X,P,ff,G,alpha0,rho,c)

% ---------- Argument defaults ----------
if ~exist('alpha0','var') | isempty(alpha0) alpha0 = 1; end;
if ~exist('rho','var') | isempty(rho) rho = 0.8; end;
if ~exist('c','var') | isempty(c) c = 1e-4; end;
% ---------- End of "argument defaults" ----------

alpha = alpha0;
tmp = c*G(:)'*P(:);
[e,sqd,esqd] = ee_error(I,J,Wp,Wn,l,X+alpha*P);
while e > ff + alpha*tmp
  alpha = rho*alpha;
  [e,sqd,esqd] = ee_error(I,J,Wp,Wn,l,X+alpha*P);
end
X = X + alpha*P;


% [e,sqd,esqd] = ee_error(I,J,Wp,Wn,l,X) EE error over X

function [e,sqd,esqd] = ee_error(I,J,Wp,Wn,l,X)

if isempty(I)
  sqd = sqdist(X); esqd = exp(-sqd);
else
  [sqd,esqd] = spsqdist(I,J,X); sqd = sqd + sqd'; esqd = esqd + esqd';
end
e = full(Wp(:)'*sqd(:) + l*Wn(:)'*esqd(:));


% [sqd,esqd] = spsqdist(I,J,T[,Y])
% Matrix of squared distances between rows of T and Y, computed only for the
% pairs in (I,J).
%
% In:
%   I,J: ?x1 lists, [I,J] are the indices of the required distances.
%   T: NxD matrix of N row D-dimensional vectors.
%   Y: MxD matrix of M row D-dimensional vectors. Default: T.
% Out:
%   sqd: NxM sparse matrix of squared Euclidean distances. sqd(n,m) is the
%      squared Euclidean distance between row vectors T(n,:) and Y(m,:).
%   esqd: exp(-sqd) where sqd is not zero.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2009 by Miguel A. Carreira-Perpinan

function [sqd,esqd] = spsqdist(I,J,T,Y)

if ~exist('Y','var') | isempty(Y)
  sqd = sparse(I,J,sum((T(I,:)-T(J,:)).^2,2),size(T,1),size(T,1));
else
  sqd = sparse(I,J,sum((T(I,:)-Y(J,:)).^2,2),size(T,1),size(Y,1));
end
esqd = spfun(@exp,-sqd);
