% [x,e,a] = ee_F(y,Y,X,l,s[,wn,method,tol,maxit]) Elastic Embedding: F mapping
%
% Maps x = F(y) for each vector y.
%
% This implementation minimises E'(x,y) over x using gradient descent with a
% good initial step size ('g') or a modified Newton method ('N').
%
% In:
%   y: MxD matrix, M D-dim data vectors.
%   Y: NxD matrix, training data vectors.
%   X: NxL matrix, latent vectors found by 'ee' for Y.
%   l: homotopy parameter used in 'ee'.
%   s: Gaussian width for the positive affinity weights used in 'ee'.
%   wn: Nx1 vector, the negative affinity weights used in 'ee'. Default: ones.
%   method: search direction method. Default: 'N'.
%   tol: minimum distance between consecutive x to keep iterating.
%      Default: 1e-5.
%   maxit: maximum number of iterations. Default: 100.
% Out:
%   x: MxL matrix, M L-dim latent vectors.
%   e, a: Mx1 cell arrays containing the sequence of errors and step sizes.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2010 by Miguel A. Carreira-Perpinan

function [x,e,a] = ee_F(y,Y,X,l,s,wn,method,tol,maxit)

% ---------- Argument defaults ----------
if ~exist('wn','var') | isempty(wn) wn = ones(size(X,1),1); end;
if ~exist('method','var') | isempty(method) method = 'N'; end;
if ~exist('tol','var') | isempty(tol) tol = 1e-5; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 100; end;
% ---------- End of "argument defaults" ----------

M = size(y,1); [N,L] = size(X); x = zeros(M,L); e = {}; a = {};

for m=1:M
  
  % Assumes positive affinities = Gaussian with width s, full graph;
  % otherwise update this line.
  sqdy = sqdist(Y,y(m,:)); Wp = exp(-sqdy/(2*s*s)); Wn_l = l*(wn.*sqdy);
  
  Wp1 = sum(Wp);
  [tmp,I] = min(sqdy); xx = X(I,:);	% Initial x
  convcrit = maxit>=1; ee = [ee_F_error(xx,X,Wp,Wn_l)]; aa = [NaN]; j = 1;
  while convcrit			% Fixed-point iteration
    xold = xx;
    dX = bsxfun(@minus,xx,X);
    Wn_l_ex2 = Wn_l.*exp(-sum(dX.^2,2)); WW = Wp - Wn_l_ex2;
    g = (4*WW)'*dX;			% Gradient
    switch method			% Search direction for each method
     case 'g'
      p = -g/(4*Wp1);			% Search direction
      %p = -g/(4*sum(WW));
     case 'N'
      s2 = sum(Wn_l_ex2);
      H = zeros(L,L); for i=1:N H = H + Wn_l_ex2(i)*dX(i,:)'*dX(i,:); end
      H = 4*((Wp1-s2)*eye(L) + 2*H);
      p = (-H\g')';			% Search direction
      if p*g' >= 0 H = H + 4*s2*eye(L); p = (-H\g')'; end;	% Safeguard
    end
    [alpha,xx,el] = ee_F_ls({X,Wp,Wn_l},xold,p,ee(end),g);	% Line search
    j = j + 1; ee(j) = el; aa(j) = alpha;
    convcrit = (norm(xx-xold) > tol*norm(xold)) & (j <= maxit);
  end
  
  x(m,:) = xx; e{m} = ee; a{m} = aa;
  
end


% [alpha,xx,el] = ee_F_ls(paramf,xx,p,ff,g[,alpha0,rho,c])
% Backtracking line search
%
% Reference: procedure 3.1, p. 41ff in:
%   Nocedal and Wright: "Numerical Optimization", Springer, 1999.
%
% In:
%   paramf: cell array containing the parameters for ee_F_error.
%   xx: 1xL vector, the current iterate.
%   p: 1xL vector, the search direction.
%   ff: value at xx of the error function.
%   g: 1xL vector, gradient at xx of the error function.
%   alpha0: initial step size. Default: 1.
%   rho: rate of decrease of the step size. Default: 0.8.
%   c: Wolfe condition. Default: 1e-4.
% Out:
%   alpha: step size.
%   xx,el: new iterate and error function value.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2010 by Miguel A. Carreira-Perpinan

function [alpha,xx,el] = ee_F_ls(paramf,xx,p,ff,g,alpha0,rho,c)

% ---------- Argument defaults ----------
if ~exist('alpha0','var') | isempty(alpha0) alpha0 = 1; end;
if ~exist('rho','var') | isempty(rho) rho = 0.8; end;
if ~exist('c','var') | isempty(c) c = 1e-4; end;
% ---------- End of "argument defaults" ----------

alpha = alpha0;
tmp = c*g*p';
el = ee_F_error(xx+alpha*p,paramf{:});
while el > ff + alpha*tmp
  alpha = rho*alpha;
  el = ee_F_error(xx+alpha*p,paramf{:});
end
xx = xx + alpha*p;


function e = ee_F_error(xx,X,Wp,Wn_l)

sqdx = sqdist(X,xx); e = 2*(Wp'*sqdx + Wn_l'*exp(-sqdx));

