% [y,e,a] = ee_f(x,Y,X,l,s[,wn,method,tol,maxit]) Elastic Embedding: f mapping
%
% Maps y = f(x) for each vector x.
%
% This implementation minimises E'(x,y) over y using gradient descent with a
% good initial step size ('g') or a modified diagonal Newton method ('N').
%
% In:
%   x: MxL matrix, M L-dim latent vectors.
% Out:
%   y: MxD matrix, M D-dim data vectors.
% Other In/Out arguments: see ee_F.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2010 by Miguel A. Carreira-Perpinan

function [y,e,a] = ee_f(x,Y,X,l,s,wn,method,tol,maxit)

% ---------- Argument defaults ----------
if ~exist('wn','var') | isempty(wn) wn = ones(size(X,1),1); end;
if ~exist('method','var') | isempty(method) method = 'N'; end;
if ~exist('tol','var') | isempty(tol) tol = 1e-5; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 100; end;
% ---------- End of "argument defaults" ----------

M = size(x,1); D = size(Y,2); y = zeros(M,D); e = {}; a = {};

for m=1:M
  
  % Assumes positive affinities = Gaussian with width s, full graph;
  % otherwise several things need modification.
  
  sqdx = sqdist(X,x(m,:)); Wn_l_x = l*(wn.*exp(-sqdx)); Wn_l_x1 = sum(Wn_l_x);
  [tmp,I] = min(sqdx); yy = Y(I,:);	% Initial y
  convcrit = maxit>=1; ee = [ee_f_error(yy,Y,s,sqdx,Wn_l_x)]; aa = [NaN]; j=1;
  while convcrit			% Fixed-point iteration
    yold = yy;
    dY = bsxfun(@minus,yy,Y); dY2 = dY.^2; sqdy = sum(dY2,2);
    Z1 = (sqdx.*exp(-sqdy/(2*s*s)))/(2*s*s); Z = -Z1 + Wn_l_x;
    g = (4*Z)'*dY;			% Gradient
    switch method			% Search direction for each method
     case 'g'
      p = -g/(4*Wn_l_x1);		% Search direction
      %p = -g/(4*sum(Z));
     case 'N'
      sZ1 = sum(Z1); H = 4*(Wn_l_x1 - sZ1 + (Z1'/(s*s))*dY2);
      if min(H)<=0 H = H + 4*sZ1; end;	% Safeguard
      p = -g./H;			% Search direction
    end
    [alpha,yy,el] = ee_f_ls({Y,s,sqdx,Wn_l_x},yold,p,ee(end),g); % Line search
    j = j + 1; ee(j) = el; aa(j) = alpha;
    convcrit = (norm(yy-yold) > tol*norm(yold)) & (j <= maxit);
  end
  
  y(m,:) = yy; e{m} = ee; a{m} = aa;
  
end


% [alpha,yy,el] = ee_f_ls(paramf,yy,p,ff,g[,alpha0,rho,c])
% Backtracking line search
%
% Reference: procedure 3.1, p. 41ff in:
%   Nocedal and Wright: "Numerical Optimization", Springer, 1999.
%
% In:
%   paramf: cell array containing the parameters for ee_f_error.
%   yy: 1xD vector, the current iterate.
%   p: 1xD vector, the search direction.
%   ff: value at yy of the error function.
%   g: 1xD vector, gradient at yy of the error function.
%   alpha0: initial step size. Default: 1.
%   rho: rate of decrease of the step size. Default: 0.8.
%   c: Wolfe condition. Default: 1e-4.
% Out:
%   alpha: step size.
%   yy,el: new iterate and error function value.
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2010 by Miguel A. Carreira-Perpinan

function [alpha,yy,el] = ee_f_ls(paramf,yy,p,ff,g,alpha0,rho,c)

% ---------- Argument defaults ----------
if ~exist('alpha0','var') | isempty(alpha0) alpha0 = 1; end;
if ~exist('rho','var') | isempty(rho) rho = 0.8; end;
if ~exist('c','var') | isempty(c) c = 1e-4; end;
% ---------- End of "argument defaults" ----------

alpha = alpha0;
tmp = c*g*p';
el = ee_f_error(yy+alpha*p,paramf{:});
while el > ff + alpha*tmp
  alpha = rho*alpha;
  el = ee_f_error(yy+alpha*p,paramf{:});
end
yy = yy + alpha*p;


function e = ee_f_error(yy,Y,s,sqdx,Wn_l_x)

sqdy = sqdist(Y,yy); e = 2*(Wn_l_x'*sqdy + sqdx'*exp(-sqdy/(2*s*s)));

